from io import StringIO
import pickle
from tempfile import NamedTemporaryFile
import uuid
import socket
import struct
from select import select
import numpy as np
import logging
import os

# Required so that sphinx can build this on readthedocs.org
try:
    from . import actxobjects
except Exception:
    pass

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
log.addHandler(logging.StreamHandler())

TYPE_NUMPY_DATA = 0
TYPE_PICKLE_DATA = 1

# Result codes
SUCCESS = 2**0
DEV_NOT_CONNECTED = 2**1
DEV_CONNECTED = 2**2
TAG_MISSING = 2**3


def _write_preamble(sock, mid, nbytes, protocol=0):
    preamble = struct.pack('!IIH', mid, nbytes, protocol)
    sock.sendall(preamble)


def _read_preamble(sock):
    mid, nbytes, protocol = struct.unpack('!IIH', sock.recv(10))


def _read(sock):
    preamble = sock.recv(10)
    if preamble == '':
        raise socket.error('transmission terminated')
    mid, size, protocol = struct.unpack('!IIH', preamble)
    log.debug('reading message id %d with %d bytes', mid, size)
    sock.settimeout(10)

    # Ensure we load the entire buffer.  This is a blocking operation and
    # probably should be optimized away.  However, this is required for for
    # reading the raw *.rcx binary data from the client.  All other methods I
    # have tested don't require this loop.
    message = StringIO()
    while True:
        message.write(sock.recv(size))
        if message.tell() == size:
            break
    return mid, protocol, message.getvalue()


def write(sock, mid, data):
    '''
    Two protocols are currently supported.  Both share a common header format.
    '''
    # If the data is a numpy array, it is much faster to simply serialize it
    # using Numpy's own protocol.  Pickle is about 4x slower.  For all other
    # datatypes, default to the Pickle protocol.
    if isinstance(data, np.ndarray):
        size = len(data.dtype.str) + len(data.data)
        _write_preamble(sock, mid, size, TYPE_NUMPY_DATA)
        sock.sendall(data.dtype.str)
        sock.sendall(data.data)
    else:
        message = pickle.dumps(data, pickle.HIGHEST_PROTOCOL)
        _write_preamble(sock, mid, len(message), TYPE_PICKLE_DATA)
        sock.sendall(message)


def read(sock):
    mid, protocol, message = _read(sock)
    if protocol == TYPE_NUMPY_DATA:
        data = np.fromstring(message[3:], dtype=message[:3])
    elif protocol == TYPE_PICKLE_DATA:
        data = pickle.loads(message)
    return mid, data


class TDTRPCServer(object):
    '''
    Remote procedure call server for communicating with TDT's System 3 hardware

    The server maintains a list of all the TDT devices that the server is
    currently communicating with.  Note that all TDT devices share the same
    connection with the computer, so we can only talk to a single device at a
    time.  Since this server processes each request sequentially rather than
    spawining each connection off to a separate thread, concurrency is handled
    by the select() loop in `run_forever`.  A thread-based design was
    considered; however, the bottleneck currently is in the optical interface
    I/O speed so it is unlikely that the additional hassle and overhead of
    threading will provide any significant performance gain.

    This is meant to be a relatively thin layer around the ActiveX device
    driver.  All messages contain a preamble (header) consisting of an integer
    (message id), half-integer (typecode) and integer (message size).  The
    message consists of three-element tuple (device_name, command_name, *args)
    where:

        device_name
            The name that you used when creating the initial connection attempt
            (e.g. the call to ConnectRZ5, Connect RX8, etc.).  Following the
            initial connection attempt, you will use that name to indicate
            which device is the target of subsequent commands.  Typically this
            name will be the device ID (e.g. 'RP2', 'RX8', 'RZ6', etc.), but
            you can certainly assign IDs such 'stim' and 'acq'.

        method_name
            The name of the command to call on the ActiveX object associated
            with the device indicated by device_name.

        args
            Tuple of arguments that will be unpacked when calling the method on
            the ActiveX object
    '''

    def __init__(self, address, connections=2, interface='GB'):
        from . import actxobjects
        self._connections = 2
        self._server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._server.bind(address)
        self._server.listen(connections)
        self._socks = [self._server]
        log.info('server running on %s:%s', *address)

        # A list of all devices that the server is connected to.  Since each
        # device has its own COM object instance, we need to maintain a map of
        # the device to the appropriate instance.
        self._iface = {}

        # Connect to the hardware
        self._interface = interface
        self._zbus = actxobjects.ZBUSx()
        self._zbus.ConnectZBUS(interface)
        # self._rpcox = actxobjects.RPcoX()
        # self._PA5 = actxobjects.PA5x()

    def run_forever(self, poll_interval=0.5):
        socks = self._socks
        server = self._server

        while True:
            # Timeout should be set to something reasonably low so that Ctrl-C
            # forces the server to exit without waiting forever.
            sread, swrite, serr = select(socks, socks, [], poll_interval)
            for sock in sread:
                if sock == server:
                    # This is an incoming connection request, let's accept it
                    # and set it up with the system.
                    self.handle_accept()
                else:
                    # This is probably a remote procedure call request from one
                    # of the clients.  Let's attempt to process the request  If
                    # an error is raised during the attempt, this most likely
                    # means the socket disconnected.
                    try:
                        self.handle_read(sock)
                    except socket.error:
                        self.handle_disconnect(sock)

    def handle_disconnect(self, sock):
        host, port = sock.getpeername()
        log.info('client from %s:%s disconnected', host, port)
        sock.close()
        self._socks.remove(sock)

    def handle_accept(self):
        new, (host, port) = self._server.accept()
        new.setblocking(0)
        self._socks.append(new)
        log.info('client from %s:%s connected', host, port)

    def handle_read(self, sock):
        from . import actxobjects
        mid, (uuid, driver, command, args) = read(sock)
        log.info('RPC request %s from %s', command, uuid)
        if command.startswith('Connect'):
            self._iface[uuid] = getattr(actxobjects, driver)()
            result = getattr(self._iface[uuid], command)(*args)
        elif command in ('LoadCOF', 'LoadCOFsf', 'ReadCOF'):
            # This is one of the TDT methods that requires a string filename.
            # Since the client is sending the data via a binary conneciton, we
            # need to convert this binary data into a temporary file whose name
            # can be passed to the method.  On platforms >= Windows NT, the
            # temporary file must be closed before it can be opened a second
            # time for reading (e.g. by the ActiveX object).  To ensure that we
            # can open it again a second time, we need to ensure that it is not
            # auto-deleted when closed.  NOTE: The *.rcx extension is required
            # for the LoadCOF method to recognize the file.
            fh = NamedTemporaryFile(suffix='.rcx', delete=False)
            fh.write(args[0])
            fh.close()
            log.info('Created temporary file %s for RCX circuit', fh.name)

            # Replace the binary file data in the argument list for the method
            # with the path to the temporary file
            args = list(args)
            args[0] = fh.name
            result = getattr(self._iface[uuid], command)(*args)
            # Now, close the temporary file we created
            os.unlink(fh.name)
        else:
            result = getattr(self._iface[uuid], command)(*args)
        write(sock, mid, result)


class _Method(object):
    '''
    Bind a RPC method to a RPC server
    '''

    def __init__(self, name, send):
        self._name = name
        self._send = send

    def __call__(self, *args):
        return self._send(self._name, *args)


class _NET(object):

    def __init__(self, address=('localhost', 13131)):
        cn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        cn.connect(address)
        self._cn = cn
        self._mid = -1
        self._results = {}
        self._uuid = str(uuid.uuid4())

    def _send(self, command, *args):
        '''
        Send command to server for execution.  Returns ID that can be used to
        look up the status of the command later.
        '''
        self._mid += 1
        write(self._cn, self._mid, (self._uuid, self.driver, command, args))
        return self._receive(self._mid)

    def _receive(self, mid, timeout=None):
        new_mid, result = read(self._cn)
        if new_mid == mid:
            return result
        else:
            raise Exception

    def __getattr__(self, name):
        return _Method(name, self._send)


class PA5NET(_NET):
    '''
    Remote procedure call client used in conjunction with the TDTRPCSever

    This is essentially a network-aware version of the PA5x object that is
    designed to be a drop-in replacement (i.e. it supports all the methods and
    attributes of the native PA5x object).  Must be used in conjunction with
    TDTRPCServer.  TDTRPCServer handles concurrency issues, allowing multiple
    threads, processes and computers to connect to the hardware simultaneously.
    '''

    driver = 'PA5x'


class zBUSNET(_NET):
    '''
    Remote procedure call client used in conjunction with the TDTRPCSever

    This is essentially a network-aware version of the zBUS object that is
    designed to be a drop-in replacement (i.e. it supports all the methods and
    attributes of the native zBUS object).  Must be used in conjunction with
    TDTRPCServer.  TDTRPCServer handles concurrency issues, allowing multiple
    threads, processes and computers to connect to the hardware simultaneously.
    '''

    driver = 'ZBUSx'


class RPcoXNET(_NET):
    '''
    Remote procedure call client used in conjunction with the TDTRPCSever

    This is essentially a network-aware version of the RPcoX object that is
    designed to be a drop-in replacement (i.e. it supports all the methods and
    attributes of the native RPcoX object).  Must be used in conjunction with
    TDTRPCServer.  TDTRPCServer handles concurrency issues, allowing multiple
    threads, processes and computers to connect to the hardware simultaneously.

    Note that a few methods may require special handling (e.g. LoadCOF) and are
    stubbed out as a result.
    '''

    driver = 'RPcoX'

    def LoadCOF(self, filename):
        '''
        Note that the legacy *.rco format is currently not supported!  You must
        send the *.rcx file.
        '''
        with open(filename, 'rb') as fh:
            return self._send('LoadCOF', fh.read())

    def LoadCOFsf(self, filename, sf):
        '''
        Note that the legacy *.rco format is currently not supported!  You must
        send the *.rcx file.
        '''
        with open(filename, 'rb') as fh:
            return self._send('LoadCOFsf', fh.read(), sf)


def test_client():
    from dsp_project import DSPProject
    project = DSPProject(address=('localhost', 13131))
    filename = ('e:/programs/development/src/neurobehavior/'
                'components/positive-behavior-v2.rcx')
    circuit = project.load_circuit(filename, 'RZ6')  # noqa
    # print circuit.tags
    # circuit.start()
    # time

    # from util import connect_rpcox, connect_zbus
    # zbus = connect_zbus(('localhost', 13131))
    # iface = connect_rpcox('RZ6', 1, ('localhost', 13131))
    # print iface.ConnectRZ6('GB', 1)
    # print iface.LoadCOF(
    # print iface.Run()
    # print iface.GetTagVal('resp_dur_n')
    # print iface.Halt()
    # client = TDTClient()
    # mid = client.execute('RZ6', 'ConnectRZ6', 'GB', 1)
    # print client.get_result(mid)
    # mid = client.execute('RZ6', 'SetTagVal', 'duration', 5)
    # print client.get_result(mid)

if __name__ == '__main__':
    import argparse

    class ParseAddress(argparse.Action):

        def __call__(self, parser, args, value, option_string=None):
            host, port = value.split(':')
            setattr(args, self.dest, (host, int(port)))

    parser = argparse.ArgumentParser(description='Run TDT System 3 RPC server')
    parser.add_argument('address', action=ParseAddress, default=('', 3333))
    args = parser.parse_args()
    TDTRPCServer(address=args.address).run_forever()



from .RPcoX import RPcoX
from .ZBUSx import ZBUSx
from .PA5x import PA5x

'''
The files in this directory were generated using makepy.  If TDT ever updates
their drivers with new methods and signatures, the files may need to be
regenerated.  To replicate the process, open up a command window and type:

>>> python -m win32com.client.makepy

A popup window with all the COM libraries available will appear.  Select the
library you want (i.e. TDevAccX) and click OK.  You'll get the following
message:

>>> python -m win32com.client.makepy
Generating to c:\Python26\lib\site-packages\win32com\gen_py\831D8AF7-7E2B-426B-A430-18E670F56C12x0x10x9.py
Building definitions from type library...
Generating...
Importing module

Pay attention to the "Generating to" path.  Find the file described and copy it
to this folder (and rename it to something more appropriate).

.. note::

    The RPcoX.py file has been modified by hand to optimize certain method calls
    (ReadTagV and ReadTagVEX).  See `Brad Buran's post` for more detail.

.. _`Brad Buran's post`:  http://bradburan.com/2011/03/speeding-up-readtagv-and-readtagvex/
'''



import numpy as np
import logging

from .abstract_ring_buffer import AbstractRingBuffer

log = logging.getLogger(__name__)


class SharedRingBuffer(AbstractRingBuffer):
    '''
    Implementation that reads/writes data to a shared memory location.  Used
    for inter-process communication.  The assumption is one process only reads
    and the other process only writes.  Does not support single-write,
    multi-read (i.e. only one process can do the reading).

    Cache *must* be a view into a Numpy array.  use shmem_to_ndarray and
    reshape.
    '''

    def __init__(self, cache, iwrite, iread, ioffset, condition, circuit):
        self._ioffset = ioffset
        self._iwrite = iwrite
        self._iread = iread
        self._cache = cache
        self._condition = condition
        self._circuit = circuit
        self._dtype = self._cache.dtype
        self.channels, self.size = self._cache.shape
        self.block_size = 1

        self._ioffset.value = -1
        self._iwrite.value = 0
        self._iread.value = 0
        self._processed = False

    def _get_empty_array(self, samples):
        return np.empty((self.channels, samples), dtype=self._dtype)

    def _get_read_index(self):
        return self._iread.value

    def _set_read_index(self, value):
        self._iread.value = int(value)

    read_index = property(_get_read_index, _set_read_index)

    def _get_write_index(self):
        return self._iwrite.value

    def _set_write_index(self, value):
        self._iwrite.value = int(value)

    write_index = property(_get_write_index, _set_write_index)

    def _read(self, offset, length):
        return self._cache[..., offset:offset+length]

    def _write(self, offset, data):
        samples = data.shape[-1]
        self._cache[..., offset:offset+samples] = data
        return samples

    def should_set(self):
        return self._ioffset.value >= 0

    # Read and write should be locked to prevent concurrent access by the two
    # proceses.  Even though one process is read-only and the other is
    # write-only, I haven't had a chance to rigorously debug what happens if we
    # don't lock access to the read/write.  The lock should be a re-entrant
    # lock.  Since set() acquires a lock and then calls write(), this allows
    # write() to acquire the same lock as well.  A simple, non-reentrant lock
    # would prevent write() from acquiring lock().
    def read(self, samples=None):
        with self._condition:
            return super(SharedRingBuffer, self).read(samples)

    def write(self, data, timeout=None):
        with self._condition:
            result = super(SharedRingBuffer, self).write(data)
            # Now, wait till the subprocess acknowledges that it has recieved
            # the data and has uploaded it to the hardware before returning. By
            # telling _condition to wait, it will release the lock, and then
            # wait for the other process to issue a _condition.notify() signal.
            # At this point, the lock returns to this thread.
            self._condition.wait(timeout)
        return result

    def set(self, data, timeout=None):
        with self._condition:
            # Locking is very important here because we need to ensure that
            # both _ioffset and the data are written before the other process
            # has access to it.
            self._ioffset.value = 0
            self.write(data, timeout=timeout)

    def clear(self):
        with self._condition:
            self._circuit.clear_buffer(self.data_tag)

    def notify(self):
        with self._condition:
            self._condition.notify()


class ReadableSharedRingBuffer(SharedRingBuffer):

    def _write(self, offset, data):
        raise NotImplementedError


class WriteableSharedRingBuffer(SharedRingBuffer):

    def _read(self, offset, data):
        raise NotImplementedError



from .util import connect_zbus
from .dsp_circuit import DSPCircuit
from .dsp_error import DSPError
import atexit

import logging
log = logging.getLogger(__name__)


class DSPProject(object):
    '''
    Used to manage loading circuits to multiple DSPs.  Mainly a convenience
    method.
    '''

    def __init__(self, address=None, interface='GB'):
        self._circuit_info = {}
        self._circuits = {}
        self._zbus = connect_zbus(interface=interface, address=address)
        self._interface = interface
        self.server_address = address
        atexit.register(self.stop)

    def load_circuit(self, circuit_name, device_name, device_id=1, **kw):
        '''
        Load the circuit to the specified device

        Parameters
        ----------
        circuit_name : str
            Path to circuit to load
        device_name : str
            Name of TDT System3 device to load circuit to
        device_id : number
            ID of device

        Returns
        -------
        circuit : instance of DSPCircuit
            The circuit.
        '''
        self._circuit_info[(circuit_name, device_name)] = []
        # We need to store a reference to the circuit here so we can properly
        # initialize any buffers we need
        circuit = DSPCircuit(circuit_name, device_name,
                             address=self.server_address,
                             interface=self._interface,
                             device_id=device_id, **kw)
        self._circuits[device_name] = circuit
        return circuit

    def start(self):
        '''
        Start all circuits that have been loaded
        '''
        for circuit in self._circuits.values():
            circuit.start()

    def stop(self):
        '''
        Stop all circuits that have been loaded
        '''
        for circuit in self._circuits.values():
            circuit.stop()

    def trigger(self, trigger, mode='pulse'):
        '''
        Fire a zBUS trigger

        Parameters
        ----------
        trigger : {'A', 'B'}
            Fire the specified trigger.  If integer, this corresponds to
            RPco.X.SoftTrg.  If 'A' or 'B', this fires the corresponding zBUS
            trigger.
        mode : {'pulse', 'high', 'low'}
            Indicates the corresponding mode to set the zBUS trigger to

        Note that due to a bug in the TDT ActiveX library for versions greater
        than 56, we have no way of ensuring that zBUS trigger A or B were
        fired.
        '''
        # Convert mode string to the corresponding integer
        mode_enum = dict(pulse=0, high=1, low=2)
        # We have no way of ensuring that zBUS trigger A or B were fired
        # properly due to a bug in versions of the ActiveX library >= 57.
        if trigger == 'A':
            self._zbus.zBusTrigA(0, mode_enum[mode], 10)
        elif trigger == 'B':
            self._zbus.zBusTrigB(0, mode_enum[mode], 10)
        else:
            mesg = "Unsupported trigger mode %s %s" % (trigger, mode)
            raise DSPError(self, mesg)
        log.debug('Trigger %r %s', trigger, mode)



'''
.. module:: tdt.dsp_buffer
    :synopsis: Module for handling I/O with the buffers on the DSP devices
.. moduleauthor:: Brad Buran <bburan@alum.mit.edu
'''
import time
import numpy as np

from .util import dtype_to_type_str, resolution
from .dsp_error import DSPError
from .constants import RCX_BUFFER
from .abstract_ring_buffer import AbstractRingBuffer

import logging
log = logging.getLogger(__name__)


class DSPBuffer(AbstractRingBuffer):
    '''
    Given the circuit object and tag name, return a buffer object that serves
    as a wrapper around a SerStore or SerSource component.  See the TDTPy
    documentation for more detail on buffers.
    '''

    # List of all attributes available.  The majority of these attributes are
    # generated by inspection of the RCX file and some may be useful for the
    # user (e.g. fs, size, etc.).

    ATTRIBUTES = [
        'data_tag', 'idx_tag', 'size_tag', 'sf_tag', 'cycle_tag',
        'dec_tag', 'src_type', 'dest_type', 'compression', 'resolution',
        'sf', 'dec_factor', 'fs', 'n_slots', 'n_samples', 'size',
        'n_slots_max', 'n_samples_max', 'size_max', 'channels',
        'block_size']

    def __init__(self, circuit, data_tag, idx_tag=None, size_tag=None,
                 sf_tag=None, cycle_tag=None, dec_tag=None, block_size=1,
                 src_type='float32', dest_type='float32', channels=1,
                 dec_factor=None, latch_trigger=None):

        if data_tag not in circuit.tags:
            raise ValueError("%s: Does not have data tag %s"
                             % (circuit, data_tag))
        elif not circuit.tags[data_tag][1] == RCX_BUFFER:
            raise ValueError("Tag %s is not a buffer tag" % data_tag)

        self.circuit = circuit
        self._iface = circuit._iface
        self._zbus = circuit._zbus
        self.data_tag = data_tag
        self.channels = channels
        self.block_size = int(block_size)
        self.latch_trigger = latch_trigger

        self.size_tag = self.find_tag(size_tag, '_n', True, 'size')
        self.idx_tag = self.find_tag(idx_tag, '_i', True, 'index')
        self.sf_tag = self.find_tag(sf_tag, '_sf', False, 'scaling factor')
        self.cycle_tag = self.find_tag(cycle_tag, '_c', False, 'cycles')
        self.dec_tag = self.find_tag(dec_tag, '_d', False, 'decimation')
        self.sf = self.get_tag(self.sf_tag, 1, 'scaling factor')

        if dec_factor is not None:
            if self.dec_tag is not None:
                self.circuit.set_tag(self.dec_tag, dec_factor)
            elif dec_factor != 1:
                m = 'Decimation tag for %s must be available to set ' \
                    'decimation factor to %d' % (self.data_tag, dec_factor)
                raise ValueError(m)

        self.dec_factor = self.get_tag(self.dec_tag, 1, 'decimation factor')
        self.fs = circuit.fs / float(self.dec_factor)
        log.debug('%s: Sampling rate %f', self, self.fs)

        # Numpy's dtype function is quite powerful and accepts a variety of
        # strings as well as other dtype objects and returns the right answer.
        self.src_type = np.dtype(src_type)
        self.dest_type = np.dtype(dest_type)

        # Number of samples compressed into a single slot.  The RPvds works
        # with 32 bit words.  If we are compressing our data, calculate the
        # number of samples that are compressed into a single 32 bit word.
        # If src_type is int8, we know we are compressing 4 samples into a
        # single 32 bit (4 byte) word.
        self.compression = int(4/np.nbytes[self.src_type])

        # Query buffer for it's size in terms of slots, samples and samples per
        # channel
        self._update_size()

        # Convert our preferred representation for the data type to TDT's
        # preferred representation for the data type.
        self.vex_src_type = dtype_to_type_str(self.src_type)
        self.vex_dest_type = dtype_to_type_str(self.dest_type)

        try:
            self.resolution = resolution(self.src_type, self.sf)
        except:
            # Does the logic change if sf is not 1?  We should never see this
            # use-case but let's add a check for safety.
            if self.sf != 1:
                raise ValueError("FIXME")
            self.resolution = np.finfo(self.src_type).resolution

        # The number of slots in the buffer must be a multiple of channel
        # number, otherwise data will be lost.  This is a requirement of the
        # RPvds circuit, so let's check to make sure this requirement is met as
        # it is a very common mistake to make.
        if (self.n_slots % self.channels) != 0:
            mesg = 'Buffer size must be a multiple of the channel number'
            raise DSPError(self, mesg)

        # Spit out debugging information
        log.debug('Initialized %s', self._get_debug_info())

    def latch(self):
        if self.latch_trigger is not None:
            self.circuit.trigger(self.latch_trigger)

    def find_tag(self, tag, default_prefix, required, name):
        '''
        Locates tag that tracks a feature of the buffer

        Parameters
        ----------
        tag : {None, str}
            Name provided by the end-user code
        default_prefix : str
            Prefix to append to the data tag name to create the default tag
            name for the feature.
        required : bool
            If the tag is required and it is missing, raise an error.
            Otherwise, return None.
        name : str
            What the tag represents. Used by the logging and exception
            machinery to create a useful message.

        Returns
        -------
        tag_name : {None, str}
            Name of tag. If no tag found and it is not required, return None.

        Raises
        ------
        ValueError
            If tag cannot be found and it is required.
        '''
        # If no name was provided create a default one.
        if tag is None:
            tag = self.data_tag + default_prefix

        # If tag exists, return it.
        if tag in self.circuit.tags:
            log.debug("%s: found %s tag %s", self, name, tag)
            return tag

        # If we have reached this point, the tag is missing. If it was
        # required, raise an error.
        if required:
            m = '%s tag for %s must be present in circuit' % (name, self)
            raise ValueError(m)
        log.debug('%s: no tag found for %s', self, name)

    def get_tag(self, tag, default, name):
        '''
        Returns value of tag that tracks a feature of the buffer

        Parameters
        ----------
        tag : {None, str}
            Name provided by the end-user code
        default : {int, float}
            Default value of feature if tag is missing.
        name : str
            What the tag represents. Used by the logging and exception
            machinery to create a useful message.

        Returns
        -------
        value : {int, float}
            Value of tag. If no tag is present, default is returned.
        '''
        if tag is None:
            log.debug('%s: %s is %r (default)', self, name, default)
            return default
        value = self.circuit.get_tag(tag)
        log.debug('%s: %s is %r', self, name, value)
        return value

    def _get_debug_info(self):
        attr_strings = ['{0}: {1}'.format(*a)
                        for a in self.attributes().items()]
        return '%s: %s' % (self, ', '.join(attr_strings))

    def attributes(self, attributes=None):
        if attributes is None:
            attributes = self.ATTRIBUTES
        return dict((attr, getattr(self, attr)) for attr in attributes)

    def __getstate__(self):
        '''
        Provides support for pickling, which is required by the multiprocessing
        module for launching a new process.  _iface is a PyIDispatch object,
        which does not support pickling, so we just delete them and pickle the
        rest.
        '''
        state = self.__dict__.copy()
        del state['_iface']
        return state

    def __setstate__(self, state):
        '''
        Loads the state and reconnects the COM objects
        '''
        self.__dict__.update(state)
        self._iface = self.circuit._iface

    def _read(self, offset, length):
        raise NotImplementedError

    def _write(self, offset, data):
        raise NotImplementedError

    def set_size(self, size):
        if not self._iface.SetTagVal(self.size_tag, size):
            raise DSPError(self, "Unable to set buffer size to %d" % size)
        self._update_size()

    def _update_size(self):
        '''
        Query current state of buffer (size and current index).  If data
        compression is being used, multiple samples can fit into a single slot
        of a RPvds buffer.  We want the index and size attributes to accurately
        reflect the number of samples per channel in the buffer, not the number
        of slots.

        For example, assume I am storing 16 channels in a buffer with two
        samples compressed into each slot.  After the DSP clock has counted 10
        samples, 10 samples per channel will have been stored for a total of
        160 samples.  However, only 80 slots will have been filled.

        n_slots
            number of slots in buffer
        n_samples
            number of samples in buffer
        size
            number of samples per channel
        '''
        # update size
        self.n_slots_max = self._iface.GetTagSize(self.data_tag)
        if self.size_tag is not None:
            self.n_slots = self.circuit.get_tag(self.size_tag)
        else:
            self.n_slots = self.n_slots_max
            log.debug("%s: no size tag available, using GetTagSize", self)

        self.n_samples = round(self.n_slots * self.compression)
        self.n_samples_max = round(self.n_slots_max * self.compression)
        self.size = round(self.n_samples / self.channels)
        self.size_max = round(self.n_samples_max / self.channels)
        self.sample_time = self.size / self.fs
        self.max_sample_time = self.size_max / self.fs

    def _get_empty_array(self, samples):
        return np.empty((self.channels, samples), dtype=self.dest_type)

    def __str__(self):
        return "{0}:{1}".format(self.circuit, self.data_tag)

    def __repr__(self):
        return "<{0}:{1}:{2}:{3}>".format(self.circuit, self.data_tag,
                                          self.write_index, self.size)

    def clear(self):
        '''
        Set buffer to zero

        Due to a bug in the TDT ActiveX library, RPco.X.ZeroTag does not work
        on certain hardware configurations.  TDT (per conversation with Chris
        Walters and Nafi Yasar) have indicated that they will not fix this bug.
        They have also indicated that they may deprecate ZeroTag in future
        versions of the ActiveX library.

        As a workaround, this method zeros out the buffer by writing a stream
        of zeros.
        '''
        zeros = np.zeros(self.n_samples)
        self._iface.WriteTagV(self.data_tag, 0, zeros)

    def _acquire(self, trigger, end_condition, samples=None, trials=1,
                 intertrial_interval=0, poll_interval=0.1, reset_read=True):
        '''
        Convenience function to handle core logic of acquisition.  Use
        `DSPBuffer.acquire` or `DSPBuffer.acquire_samples` instead.
        '''
        acquired_data = []
        for i in range(trials):
            if reset_read:
                self.reset_read(0)
            samples_acquired = 0
            time.sleep(intertrial_interval)
            trial_data = []
            self.circuit.trigger(trigger)
            while True:
                if end_condition(self, samples_acquired):
                    break
                new_data = self.read()
                samples_acquired += new_data.shape[-1]
                trial_data.append(new_data)
                log.debug('%s: acquired %d samples', self, samples_acquired)
                time.sleep(poll_interval)
            remaining_data = self.read()
            samples_acquired += remaining_data.shape[-1]
            trial_data.append(remaining_data)
            trial_data = np.hstack(trial_data)
            if samples is not None:
                trial_data = trial_data[:, :samples]
            acquired_data.append(trial_data[np.newaxis])
        return np.vstack(acquired_data)

    def acquire(self, trigger, handshake_tag, end_condition=None, trials=1,
                intertrial_interval=0, poll_interval=0.1, reset_read=True):
        '''
        Fire trigger and acquire resulting block of data

        Data will be continuously spooled while the status of the handshake_tag
        is being monitored, so a single acquisition block can be larger than
        the size of the buffer; however, be sure to set poll_interval to a
        duration that is sufficient to to download data before it is
        overwritten.

        Parameters
        ----------
        trigger
            Trigger that starts data acquistion (can be A, B, or 1-9)
        handshake_tag
            Tag indicating status of data acquisition
        end_condition
            If None, any change to the value of handshake_tag after trigger is
            fired indicates data acquisition is complete.  Otherwise, data
            acquisition is done when the value of handshake_tag equals the
            end_condition.  end_condition may be a Python callable that takes
            the value of the handshake tag and returns a boolean indicating
            whether acquisition is complete or not.
        trials
            Number of trials to collect
        intertrial_interval
            Time to pause in between trials
        poll_interval
            Time to pause in between polling hardware
        reset_read
            Should the read index be reset at the beginning of each acquisition
            sweep?  If data is written starting at the first index of the
            buffer, then this should be True.  If data is written continuously
            to the buffer with no reset of the index in between sweeps, then
            this should be False.

        Returns
        -------
        acquired_trials : ndarray
            A 3-dimensional array in the format (trial, channel, sample).

        Examples
        --------
        >>> buffer.acquire(1, 'sweep_done')
        >>> buffer.acquire(1, 'sweep_done', True)
        '''
        # TODO: should we set the read index to = write index?

        if end_condition is None:
            handshake_value = self.circuit.get_tag(handshake_tag)

            def is_done(x):
                return x != handshake_value
        elif not callable(end_condition):
            def is_done(x):
                return x == end_condition
        else:
            is_done = end_condition

        def wrapper(dsp_buffer, samples):
            current_value = dsp_buffer.circuit.get_tag(handshake_tag)
            return is_done(current_value)

        return self._acquire(
            trigger, end_condition=wrapper, samples=None,
            trials=trials, intertrial_interval=intertrial_interval,
            poll_interval=poll_interval, reset_read=reset_read)

    def acquire_samples(self, trigger, samples, trials=1,
                        intertrial_interval=0, poll_interval=0.1,
                        reset_read=True):
        '''
        Fire trigger and acquire n samples
        '''
        if samples % self.block_size:
            raise ValueError("Number of samples must be a multiple of "
                             "block size")
        log.debug('%s: attempting to acquire %d samples', self, samples)

        def is_done(b, s):
            return s >= samples
        return self._acquire(
            trigger, end_condition=is_done, samples=samples,
            trials=trials, intertrial_interval=intertrial_interval,
            poll_interval=poll_interval, reset_read=reset_read)


class ReadableDSPBuffer(DSPBuffer):

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.read_index = 0
        self.read_cycle = 0
        self.total_samples_read = 0

    def _get_write_index(self):
        index = int(self.circuit.get_tag(self.idx_tag))
        actual_index = index * self.compression / self.channels
        log.debug("%s: index raw %d, actual %d", self, index, actual_index)
        return actual_index

    write_index = property(_get_write_index)

    def _get_write_cycle(self):
        if self.cycle_tag is None:
            return None
        return int(self.circuit.get_tag(self.cycle_tag))

    write_cycle = property(_get_write_cycle)

    def _read(self, offset, length):
        log.debug("%s: read offset %d, read size %d", self, offset, length)
        if length == 0:
            data = np.array([], dtype=self.dest_type)
            return data.reshape((self.channels, -1))

        # At this point, we have already done the necessary computation of
        # offset and read size so all we have to do is pass those values
        # directly to the ReadTagVEX function.
        data = self._iface.ReadTagVEX(
            self.data_tag, offset, length,
            self.vex_src_type, self.vex_dest_type, self.channels)
        return np.divide(data, self.sf).astype(self.dest_type)


class WriteableDSPBuffer(DSPBuffer):

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.write_index = 0
        self.write_cycle = 0
        self.total_samples_written = 0

    def _get_read_index(self):
        # This returns the current sample that's been read
        index = int(self.circuit.get_tag(self.idx_tag))
        return index * self.compression / self.channels

    read_index = property(_get_read_index)

    def _get_read_cycle(self):
        if self.cycle_tag is None:
            return None
        return self.circuit.get_tag(self.cycle_tag)

    read_cycle = property(_get_read_cycle)

    def _write(self, offset, data):
        log.debug("%s: write %d samples at %d", self, len(data), offset)
        return self._iface.WriteTagV(self.data_tag, offset, data)

    def set(self, data):
        '''
        Assumes data is written starting at the first index of the buffer. Use
        for epoch-based playout.
        '''
        data = np.asarray(data)
        size = data.shape[-1]
        if size > self.size_max:
            mesg = "Cannot write %d samples to buffer" % size
            raise DSPError(self, mesg)
        if self.size_tag is not None:
            self.set_size(size)
        elif size != self.size:
            mesg = "Buffer size cannot be configured"
            raise DSPError(self, mesg)

        if size == 0:
            return

        if self.vex_src_type != 'F32':
            raise NotImplementedError
        if not self._iface.WriteTagV(self.data_tag, 0, data):
            raise DSPError(self, "write failed")
        log.debug("%s: set buffer with %d samples", self, size)
        self.total_samples_written += size



import pytest
import time

import numpy as np

from tdt import DSPProject


BASE_FS = 97656.25


@pytest.fixture
def project():
    return DSPProject()


@pytest.fixture
def circuit(project):
    circuit = project.load_circuit('RZ6-debugging.rcx', 'RZ6', 1, latch_trigger=1)
    circuit.start()
    yield circuit
    circuit.stop()


@pytest.fixture
def ao1(circuit):
    return circuit.get_buffer('ao1', 'w')


@pytest.fixture
def ao2(circuit):
    return circuit.get_buffer('ao2', 'w')


@pytest.fixture
def ai1(circuit):
    return circuit.get_buffer('ai1', 'r')


@pytest.fixture
def ai2(circuit):
    return circuit.get_buffer('ai2', 'r')


def test_circuit_load(circuit):
    circuit.start()
    circuit.stop()


def test_buffer_fs(ao1, ai1):
    assert ao1.fs == BASE_FS
    assert ai1.fs == BASE_FS


def test_circuit_write_read(project, ao1, ai1):
    n = round(ao1.fs)
    write_samples = np.random.uniform(size=n)
    ao1.write(write_samples)
    project.trigger('A', 'high')
    time.sleep(1)

    # There is a two-sample delay in the circuit
    read_samples = ai1.read()[0, :n]
    np.testing.assert_allclose(write_samples, read_samples)


def test_buffer_detect_dec_factor(circuit):
    # The default decimation factor is 8 by default
    ai_dec1 = circuit.get_buffer('ai_dec1', 'r')
    assert ai_dec1.fs == (BASE_FS / 8)


@pytest.mark.parametrize("dec_factor", [1, 2, 4, 8])
def test_circuit_write_read_dec(project, circuit, ao1, dec_factor):
    ai_dec = circuit.get_buffer('ai_dec1', 'r', dec_factor=dec_factor)
    assert ai_dec.fs == (BASE_FS / dec_factor)

    n = round(ao1.fs / 10)
    t = np.arange(n) / ao1.fs
    write_samples = np.sin(2 * np.pi * 50 * t)
    ao1.write(write_samples)

    project.trigger('A', 'high')
    read_samples_dec = []
    for i in range(10):
        time.sleep(0.05)
        read_samples_dec.append(ai_dec.read())

    n_dec = int(n / dec_factor)
    read_samples_dec = np.concatenate(read_samples_dec, axis=-1)[0, :n_dec]
    write_samples_dec = write_samples[::dec_factor][:n_dec]
    np.testing.assert_allclose(write_samples_dec, read_samples_dec)


def test_circuit_write_too_slow(project, ao1):
    # This tests the condition where a subsequent write takes so long that the
    # samples generated by the output buffer run into an area of the ring
    # buffer that has not received new data yet.
    n = round(ao1.fs)
    ao1.write(np.random.uniform(size=n))
    project.trigger('A', 'high')
    time.sleep(1)
    with pytest.raises(IOError, match='Write was too slow and old samples were regenerated'):
        ao1.write(np.random.uniform(size=n))


def test_circuit_missing_initial_write(project, ao1):
    # This tests the condition where the first write to the output buffer
    # occurs after the circuit is started. This means that the buffer is not
    # properly initialized and is generating random samples (whatever existed
    # when the buffer memory was allocated)
    n = round(ao1.fs)
    project.trigger('A', 'high')
    with pytest.raises(IOError, match='Write was too slow and old samples were regenerated'):
        ao1.write(np.random.uniform(size=n))


def test_circuit_read_too_slow(project, ai1):
    n = round(ai1.fs)
    ai1.set_size(n)
    project.trigger('A', 'high')
    time.sleep(1)
    with pytest.raises(IOError, match='Read was too slow and unread samples were overwritten'):
        ai1.read()


def test_circuit_read_wrap(project, ai1, ai2):
    project.trigger('A', 'high')
    d1 = []
    d2 = []
    for i in range(15):
        time.sleep(0.1)
        d1.append(ai1.read())
        d2.append(ai2.read())
    project.trigger('A', 'low')
    d1.append(ai1.read())
    d2.append(ai2.read())
    d1 = np.concatenate(d1, axis=1)
    d2 = np.concatenate(d2, axis=1)
    assert d1.shape[-1] == ((ai1.read_cycle * ai1.n_slots) + ai1.read_index)
    assert d2.shape[-1] == ((ai2.read_cycle * ai2.n_slots) + ai2.read_index)
    assert d1.shape == d2.shape
    assert ai1.read_cycle != 0


@pytest.mark.parametrize("dec_factor", [1, 2, 4, 8])
def test_circuit_dec_read_wrap(project, circuit, dec_factor):
    # Verify that decimated buffer reads properly wrap around
    ai1 = circuit.get_buffer('ai_dec1', 'r', dec_factor=dec_factor)
    read_time = ai1.n_samples / ai1.fs
    print(read_time)
    project.trigger('A', 'high')
    d1 = []
    for i in range(15):
        time.sleep(read_time * 0.1)
        d1.append(ai1.read())
    project.trigger('A', 'low')
    d1.append(ai1.read())
    d1 = np.concatenate(d1, axis=1)
    assert d1.shape[-1] == ((ai1.read_cycle * ai1.n_slots) + ai1.read_index)
    assert ai1.read_cycle != 0


def test_circuit_write_wrap(project, ao1, ao2, ai1, ai2):
    d1 = np.random.uniform(size=ao1.available())
    ao1.write(d1)
    w1 = [d1]

    d2 = np.random.uniform(size=ao2.available())
    ao2.write(d2)
    w2 = [d2]
    r1 = []
    r2 = []

    project.trigger('A', 'high')
    for i in range(15):
        time.sleep(0.1)
        d1 = np.random.uniform(size=ao1.available())
        ao1.write(d1)
        w1.append(d1)
        d2 = np.random.uniform(size=ao2.available())
        ao2.write(d2)
        w2.append(d2)

        r1.append(ai1.read())
        r2.append(ai2.read())

    project.trigger('A', 'low')
    d1 = np.random.uniform(size=ao1.available())
    ao1.write(d1)
    w1.append(d1)
    d2 = np.random.uniform(size=ao2.available())
    ao2.write(d2)
    w2.append(d2)
    r1.append(ai1.read())
    r2.append(ai2.read())

    w1 = np.concatenate(w1, axis=-1)
    w2 = np.concatenate(w2, axis=-1)
    r1 = np.concatenate(r1, axis=-1)[0]
    r2 = np.concatenate(r2, axis=-1)[0]

    assert r1.shape == r2.shape
    n = r1.shape[0]

    np.testing.assert_array_almost_equal(w1[:n], r1)
    np.testing.assert_array_almost_equal(w2[:n], r2)
